from collections import defaultdict
from contextlib import suppress
from functools import wraps
from inspect import signature
from types import new_class
from typing import (
    Callable,
    Dict,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from apischema.conversions.utils import (
    Conversions,
    ConverterWithConversions,
    check_converter,
    check_convertible,
    handle_generic_conversions,
)
from apischema.types import AnyType, OrderedDict

_deserializers: Dict[AnyType, Dict[AnyType, ConverterWithConversions]] = defaultdict(
    OrderedDict
)
_extra_deserializers: Dict[
    AnyType, Dict[AnyType, ConverterWithConversions]
] = defaultdict(OrderedDict)
_serializers: Dict[Type, Tuple[Type, ConverterWithConversions]] = {}
_extra_serializers: Dict[Type, Dict[Type, ConverterWithConversions]] = defaultdict(dict)

Func = TypeVar("Func", bound=Callable)
_ConverterSetter = Callable[[Func, AnyType, AnyType, Optional[Conversions], bool], Func]


class MethodConverter:
    def __init__(
        self,
        decorator: _ConverterSetter,
        method: Union[Callable, classmethod, staticmethod],
        conversions: Optional[Conversions],
        extra: bool,
        instance_method: bool,
    ):
        self.decorator = decorator
        self.method = method
        self.conversions = conversions
        self.extra = extra
        self.instance_method = instance_method

    def __call__(self, *args, **kwargs):
        raise RuntimeError(
            f"Converter method {self.method} __set_name__ has not been called"
        )

    @staticmethod
    def _return(owner: Type) -> Optional[Type]:
        return None

    def __set_name__(self, owner, name):
        converter = self.method.__get__(None, owner)
        if self.instance_method:
            converter = wraps(converter)(lambda instance: getattr(instance, name)())
        param = owner if self.instance_method else None
        param, ret = check_converter(
            converter, param, self._return(owner), {owner.__name__: owner}
        )
        self.decorator(converter, param, ret, self.conversions, self.extra)
        setattr(owner, name, self.method)


try:
    from apischema.typing import Protocol

    class ConverterSetter(Protocol[Func]):
        @overload
        def __call__(
            self,
            function: Func,
            param: AnyType = None,
            ret: AnyType = None,
            conversions: Conversions = None,
        ) -> Func:
            ...

        @overload
        def __call__(self, *, conversions: Conversions) -> Callable[[Func], Func]:
            ...

        def __call__(
            self,
            function: Func = None,
            param: AnyType = None,
            ret: AnyType = None,
            conversions: Conversions = None,
        ):
            ...


except ImportError:
    ConverterSetter = Callable  # type: ignore


def _converter(decorator: _ConverterSetter, *, extra: bool) -> ConverterSetter:
    def wrapper(
        func: Callable = None,
        param: AnyType = None,
        ret: AnyType = None,
        conversions: Conversions = None,
    ):
        if func is None:
            return lambda func: wrapper(func, param, ret, conversions)
        if isinstance(func, (classmethod, staticmethod)):
            return MethodConverter(
                decorator, func, conversions, extra, instance_method=False
            )
        if param is None:
            with suppress(ValueError):  # builtin type has no signature
                first_param = next(iter(signature(func).parameters.values()), None)
                if first_param is not None and first_param.name == "self":
                    try:
                        param, ret = check_converter(func, param, ret)
                        return decorator(func, param, ret, conversions, extra)
                    except Exception:  # param is not annotated or recursive
                        return MethodConverter(
                            decorator, func, conversions, extra, instance_method=True
                        )
        param, ret = check_converter(func, param, ret)
        return decorator(func, param, ret, conversions, extra)

    return cast(ConverterSetter, wrapper)


def _deserializer(
    function: Func,
    param: AnyType,
    ret: AnyType,
    conversions: Optional[Conversions],
    extra: bool,
) -> Func:
    ret, param = handle_generic_conversions(ret, param)
    check_convertible(ret)
    if param == ret:
        raise ValueError("Use self_deserializer")
    _extra_deserializers[ret][param] = function, conversions
    if not extra:
        _deserializers[ret][param] = function, conversions
    return function


deserializer = _converter(_deserializer, extra=False)
extra_deserializer = _converter(_deserializer, extra=True)


def _serializer(
    function: Func,
    param: Type,
    ret: AnyType,
    conversions: Optional[Conversions],
    extra: bool,
) -> Func:
    param, ret = handle_generic_conversions(param, ret)
    check_convertible(param)
    if param == ret:
        if conversions:
            raise ValueError("Self-conversions cannot have conversions parameter")
        if not extra:
            _serializers.pop(param, ...)
    else:
        _extra_serializers[param][ret] = (function, conversions)
        if not extra:
            _serializers[param] = ret, (function, conversions)
    return function


serializer = _converter(_serializer, extra=False)
extra_serializer = _converter(_serializer, extra=True)


Cls = TypeVar("Cls", bound=Type)


def self_deserializer(cls: Cls) -> Cls:
    _deserializer(lambda x: x, new_class(cls.__name__, (cls,)), cls, None, extra=False)
    return cls


def reset_deserializers(cls: AnyType):
    _deserializers.pop(cls, ...)


class InheritedDeserializer(MethodConverter):
    def __init__(
        self, method: classmethod, conversions: Optional[Conversions], extra: bool
    ):
        super().__init__(_deserializer, method, conversions, extra, False)

    @staticmethod
    def _return(owner: Type) -> Optional[Type]:
        return owner

    def __set_name__(self, owner, name):
        prev_init_subclass = owner.__init_subclass__

        def init_subclass(cls, **kwargs):
            prev_init_subclass(**kwargs)
            deserializer(getattr(cls, name), None, cls)

        owner.__init_subclass__ = classmethod(init_subclass)
        super().__set_name__(owner, name)


ClsMethod = TypeVar("ClsMethod")


@overload
def inherited_deserializer(func: ClsMethod) -> ClsMethod:
    ...


@overload
def inherited_deserializer(
    *, conversions: Conversions = None, extra: bool = False
) -> Callable[[ClsMethod], ClsMethod]:
    ...


def inherited_deserializer(
    func=None, *, conversions: Conversions = None, extra: bool = False
):
    if func is None:
        return lambda func: inherited_deserializer(  # type: ignore
            func, conversions=conversions
        )
    if not isinstance(func, classmethod):
        raise TypeError("inherited_deserializer must be called on classmethod")
    return InheritedDeserializer(func, conversions, extra)



from dataclasses import dataclass, is_dataclass
from types import new_class
from typing import Callable, Type, Union

from apischema.cache import cache
from apischema.conversions.converters import (
    deserializer,
    extra_deserializer,
    extra_serializer,
    serializer,
)
from apischema.conversions.utils import identity
from apischema.utils import PREFIX

MODEL_ORIGIN_ATTR = f"{PREFIX}apischema"


Model = Union[Type, Callable[[], Type]]


@dataclass(frozen=True)
class DataclassModelWrapper:
    cls: Type
    model: Model


@cache
def get_model(cls: Type, model: Model) -> Type:
    if not isinstance(model, type):
        model = model()

    if not is_dataclass(model):
        raise TypeError("Dataclass model must be a dataclass")
    if len(getattr(cls, "__parameters__", ())) != len(
        getattr(model, "__parameters__", ())
    ):
        raise TypeError("Dataclass model must have the same generic parameters")
    base = model[cls.__parameters__] if hasattr(cls, "__parameters__") else model

    def __new__(_, *args, **kwargs):
        return cls(*args, **kwargs)

    return new_class(
        model.__name__,
        (base,),
        {},
        lambda ns: ns.update({"__new__": __new__, MODEL_ORIGIN_ATTR: cls}),
    )


def dataclass_model(
    cls: Type, *, deserialization=True, serialization=True, extra=False
) -> Callable[[Model], DataclassModelWrapper]:
    def decorator(model: Model) -> DataclassModelWrapper:
        wrapped_model = DataclassModelWrapper(cls, model)
        if deserialization:
            if extra:
                extra_deserializer(identity, wrapped_model, cls)
            else:
                deserializer(identity, wrapped_model, cls)
        if serialization:
            if extra:
                extra_serializer(identity, cls, wrapped_model)
            else:
                serializer(identity, cls, wrapped_model)

        return wrapped_model

    return decorator


def get_model_origin(cls: Type) -> Type:
    return getattr(cls, MODEL_ORIGIN_ATTR, cls)



from dataclasses import Field
from enum import Enum
from typing import (  # type: ignore
    Any,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
)

from apischema.json_schema.generation.visitor import SchemaVisitor
from apischema.json_schema.refs import get_ref, schema_ref
from apischema.types import AnyType
from apischema.utils import is_hashable

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

Refs = Dict[str, Tuple[AnyType, int]]


class Recursive(Exception):
    pass


class RefsExtractor(SchemaVisitor):
    def __init__(self, refs: Refs):
        super().__init__()
        self.refs = refs

    def _incr_ref(self, ref: Optional[str], cls: AnyType) -> bool:
        if ref is None:
            return False
        else:
            ref_cls, count = self.refs.get(ref, (cls, 0))
            if ref_cls != cls:
                raise ValueError(
                    f"Types {cls} and {self.refs[ref]} share same reference '{ref}'"
                )
            self.refs[ref] = (ref_cls, count + 1)
            return count > 0

    def annotated(self, cls: AnyType, annotations: Sequence[Any]):
        for annotation in reversed(annotations):
            if isinstance(annotation, schema_ref):
                annotation.check_type(cls)
                ref = annotation.ref
                if not isinstance(ref, str):
                    raise ValueError("Annotated schema_ref can only be str")
                annotated = Annotated[(cls, *filter(is_hashable, annotations))]  # type: ignore # noqa: E501
                if self._incr_ref(ref, annotated):
                    return
        return self.visit(cls)

    def any(self):
        pass

    def collection(self, cls: Type[Iterable], value_type: AnyType):
        return self.visit(value_type)

    def dataclass(
        self,
        cls: Type,
        types: Mapping[str, AnyType],
        fields: Sequence[Field],
        init_vars: Sequence[Field],
    ):
        for field in self._dataclass_fields(fields, init_vars):
            self._visit_field(field, types[field.name])

    def enum(self, cls: Type[Enum]):
        pass

    def literal(self, values: Sequence[Any]):
        pass

    def mapping(self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType):
        self.visit(key_type)
        self.visit(value_type)

    def named_tuple(
        self,
        cls: Type[Tuple],
        types: Mapping[str, AnyType],
        defaults: Mapping[str, Any],
    ):
        for cls in types.values():
            self.visit(cls)

    def new_type(self, cls: AnyType, super_type: AnyType):
        self.visit(super_type)

    def primitive(self, cls: Type):
        pass

    def subprimitive(self, cls: Type, superclass: Type):
        self.visit(superclass)

    def tuple(self, types: Sequence[AnyType]):
        for cls in types:
            self.visit(cls)

    def typed_dict(self, cls: Type, keys: Mapping[str, AnyType], total: bool):
        for cls in keys.values():
            self.visit(cls)

    def _union_result(self, results: Iterable):
        for _ in results:
            pass

    def visit(self, cls: AnyType):
        if (
            not is_hashable(cls)
            or self.is_extra_conversions(cls)
            or not self._incr_ref(get_ref(cls), cls)
        ):
            super().visit(cls)



from collections import ChainMap, defaultdict
from dataclasses import dataclass
from inspect import Parameter, iscoroutinefunction, signature
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    overload,
)

from apischema.conversions import Conversions
from apischema.conversions.dataclass_model import get_model_origin
from apischema.json_schema.schema import Schema
from apischema.types import AnyType
from apischema.typing import get_args, get_origin, get_type_hints
from apischema.utils import cached_property, get_origin_or_class

Description = Union[str, "ellipsis", None]  # noqa: F821

awaitable_origin = get_origin(Awaitable[Any])


@dataclass(frozen=True)
class Resolver:
    func: Callable
    wrapper: Callable
    parameters: Sequence[Parameter]
    conversions: Optional[Conversions] = None
    schema: Optional[Schema] = None

    @cached_property
    def types(self) -> Mapping[str, AnyType]:
        return get_type_hints(self.func, include_extras=True)

    @property
    def return_type(self) -> AnyType:
        ret = self.types["return"]
        return get_args(ret)[0] if get_origin(ret) == awaitable_origin else ret

    @property
    def is_async(self) -> bool:
        if iscoroutinefunction(self.func):
            return True
        try:
            return issubclass(get_origin_or_class(self.types["return"]), Awaitable)
        except Exception:  # py36 has weird AttributeError with issubclass
            return False


class MissingFirstParameter(Exception):
    pass


def resolver_parameters(resolver: Callable, *, skip_first: bool) -> Sequence[Parameter]:
    if "return" not in resolver.__annotations__:
        raise TypeError("Resolver must be typed")
    parameters = []
    params = list(signature(resolver).parameters.values())
    if skip_first:
        if not params:
            raise MissingFirstParameter
        params = params[1:]
    for param in params:
        if param.kind is Parameter.POSITIONAL_ONLY:
            raise TypeError("Resolver can not have positional only parameters")
        if param.kind in {Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY}:
            if param.annotation is Parameter.empty:
                raise TypeError("Resolver must be typed")
            parameters.append(param)
    return parameters


_resolvers: Dict[Type, Dict[str, Resolver]] = defaultdict(dict)


def get_resolvers(cls: Type) -> Mapping[str, Resolver]:
    all_resolvers = (
        _resolvers[sub_cls]
        for cls2 in (cls, get_model_origin(cls))
        for sub_cls in reversed(cls2.__mro__)
    )
    return ChainMap(*all_resolvers)


class ResolverDescriptor:
    def __init__(
        self,
        func: Any,
        name: str = None,
        conversions: Conversions = None,
        schema: Schema = None,
    ):
        self.func = func
        self.name = name
        self.conversions = conversions
        self.schema = schema

    def __set_name__(self, owner, name):
        wrapper: Callable
        if isinstance(self.func, property):
            method = self.func.fget

            def wrapper(self):
                return getattr(self, name)

        else:
            method = self.func.__get__(None, owner)

            def wrapper(self, *args, **kwargs):
                return getattr(self, name)(*args, **kwargs)

        parameters = resolver_parameters(
            method, skip_first=not isinstance(self.func, staticmethod)
        )

        _resolvers[owner][self.name or name] = Resolver(
            method,
            wrapper,
            parameters,
            self.conversions,
            self.schema,
        )
        setattr(owner, name, self.func)

    def __call__(self, *args, **kwargs):
        raise TypeError("Resolver method {self.__set_name__ has not been called")


MethodOrProperty = TypeVar(
    "MethodOrProperty", bound=Union[Callable, staticmethod, classmethod, property]
)


@overload
def resolver(__method_or_property: MethodOrProperty) -> MethodOrProperty:
    ...


@overload
def resolver(
    alias: str = None, *, conversions: Conversions = None, schema: Schema = None
) -> Callable[[MethodOrProperty], MethodOrProperty]:
    ...


def resolver(
    __arg=None,
    *,
    alias: str = None,
    conversions: Conversions = None,
    schema: Schema = None,
):
    if isinstance(__arg, str) or __arg is None:
        return lambda method: ResolverDescriptor(
            method, alias or __arg, conversions, schema
        )
    return ResolverDescriptor(__arg)


Func = TypeVar("Func", bound=Callable)


def add_resolver(
    cls: Type,
    name: str = None,
    *,
    conversions: Conversions = None,
    schema: Schema = None,
) -> Callable[[Func], Func]:
    def decorator(func: Func) -> Func:
        parameters = resolver_parameters(func, skip_first=False)
        types = get_type_hints(func)
        if get_origin_or_class(types[parameters[0].name]) == cls:
            parameters = parameters[1:]
            wrapper: Callable = func
        else:
            wrapper = lambda __, *args, **kwargs: func(*args, **kwargs)  # noqa: E731
        _resolvers[cls][name or func.__name__] = Resolver(
            func, wrapper, parameters, conversions, schema
        )
        return func

    return decorator



__all__ = [
    "Conversions",
    "Deserialization",
    "Serialization",
    "deserializer",
    "extra_deserializer",
    "extra_serializer",
    "identity",
    "inherited_deserializer",
    "raw_deserializer",
    "reset_deserializers",
    "self_deserializer",
    "serializer",
]

from .converters import (
    deserializer,
    extra_deserializer,
    extra_serializer,
    inherited_deserializer,
    reset_deserializers,
    self_deserializer,
    serializer,
)
from .raw import raw_deserializer
from .utils import Conversions, identity
from .visitor import Deserialization, Serialization



__all__ = ["add_resolver", "graphql_schema", "interface", "resolver"]

from .interfaces import interface
from apischema.resolvers import add_resolver, resolver

try:
    from .builder import graphql_schema
except ImportError:
    raise ImportError(
        "GraphQL feature requires graphql-core library\n"
        "Run `pip install apischema[graphql]` to install it"
    )



from dataclasses import *  # noqa
from typing import TypeVar

T = TypeVar("T")


def replace(__obj: T, **changes) -> T:  # type: ignore
    from apischema.fields import FIELDS_SET_ATTR, fields_set, set_fields
    from dataclasses import replace as replace_

    result = replace_(__obj, **changes)
    if hasattr(__obj, FIELDS_SET_ATTR):
        set_fields(result, *fields_set(__obj), *changes, overwrite=True)
    return result


del TypeVar, T



"""Kind of typing_extensions for this package"""
__all__ = ["_LiteralMeta", "_TypedDictMeta", "get_args", "get_origin", "get_type_hints"]

import sys
from types import ModuleType
from typing import Any, Callable, Generic, Tuple, TypeVar


class _FakeType:
    pass


if sys.version_info >= (3, 9):  # pragma: no cover
    from typing import Annotated, get_type_hints, get_origin, get_args
else:  # pragma: no cover
    try:
        from typing_extensions import Annotated
    except ImportError:
        pass
    try:
        from typing_extensions import get_type_hints as gth
    except ImportError:
        from typing import get_type_hints as _gth

        def gth(obj, globalns=None, localns=None, include_extras=False):  # type: ignore
            return _gth(obj, globalns, localns)

    def get_type_hints(  # type: ignore
        obj, globalns=None, localns=None, include_extras=False
    ):
        # TODO This has been fixed in recent 3.7 and 3.8
        # fix https://bugs.python.org/issue37838
        if not isinstance(obj, (type, ModuleType)) and globalns is None:
            nsobj = obj
            while hasattr(nsobj, "__wrapped__"):
                nsobj = nsobj.__wrapped__
            globalns = getattr(nsobj, "__globals__", None)
        localns = {"unicode": str, **(localns or {})}
        return gth(obj, globalns, localns, include_extras)

    try:
        from typing_extensions import get_origin, get_args
    except ImportError:

        def _assemble_tree(tree: Tuple[Any]) -> Any:
            if not isinstance(tree, tuple):
                return tree
            else:
                origin, *args = tree  # type: ignore
                if origin is Annotated:
                    return Annotated[(_assemble_tree(args[0]), *args[1])]
                else:
                    return origin[tuple(map(_assemble_tree, args))]

        def get_origin(tp):  # type: ignore
            # In Python 3.6: List[Collection[T]][int]._args__ == int != Collection[int]
            if hasattr(tp, "_subs_tree"):
                tp = _assemble_tree(tp._subs_tree())
            if isinstance(tp, _AnnotatedAlias):
                return Annotated
            if tp is Generic:
                return Generic
            return getattr(tp, "__origin__", None)

        def get_args(tp):  # type: ignore
            # In Python 3.6: List[Collection[T]][int]._args__ == int != Collection[int]
            if hasattr(tp, "_subs_tree"):
                tp = _assemble_tree(tp._subs_tree())
            if isinstance(tp, _AnnotatedAlias):
                return (tp.__args__[0], *tp.__metadata__)
            res = getattr(tp, "__args__", ())
            if get_origin(tp) is Callable and res[0] is not Ellipsis:
                res = (list(res[:-1]), res[-1])
            return res


if sys.version_info >= (3, 8):  # pragma: no cover
    from typing import Literal, TypedDict, Protocol  # noqa: F401
else:  # pragma: no cover
    try:
        from typing_extensions import Literal, TypedDict, Protocol  # noqa: F401
    except ImportError:
        pass


_T = TypeVar("_T")
_GenericAlias: Any = type(Generic[_T])
try:
    _AnnotatedAlias: Any = type(Annotated[_T, ...])
except NameError:
    _AnnotatedAlias = _FakeType
try:

    class _TypedDictImplem(TypedDict):
        pass

    _LiteralMeta: Any = type(Literal)
    _TypedDictMeta: Any = type(_TypedDictImplem)
except NameError:
    _LiteralMeta, _TypedDictMeta = _FakeType, _FakeType  # type: ignore



from enum import Enum, auto
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from apischema.types import AnyType
from apischema.typing import _GenericAlias, get_origin

PREFIX = "_apischema_"


# Singleton type, see https://www.python.org/dev/peps/pep-0484/#id30
class UndefinedType(Enum):
    def __repr__(self):
        return "Undefined"

    def __str__(self):
        return "Undefined"

    def __bool__(self):
        return False

    Undefined = auto()


Undefined = UndefinedType.Undefined


def is_hashable(obj) -> bool:
    try:
        hash(obj)
    except Exception:  # should be TypeError, but who knows what can happen
        return False
    else:
        return True


def to_hashable(data: Union[None, int, float, str, bool, list, dict]) -> Hashable:
    if isinstance(data, list):
        return tuple(map(to_hashable, data))
    if isinstance(data, dict):
        return tuple(sorted((to_hashable(k), to_hashable(v)) for k, v in data.items()))
    return data  # type: ignore


def to_camel_case(s: str):
    pascal_case = "".join(map(str.capitalize, s.split("_")))
    return pascal_case[0].lower() + pascal_case[1:]


_type_hints: Dict[str, Mapping[str, Type]] = {}


def type_name(cls: AnyType) -> str:
    if hasattr(cls, "__name__"):
        return cls.__name__
    elif isinstance(cls, _GenericAlias):
        return cls._name
    else:
        raise NotImplementedError()


MakeDataclassField = Union[Tuple[str, AnyType], Tuple[str, AnyType, Any]]

T = TypeVar("T")


def merge_opts(
    func: Callable[[T, T], T]
) -> Callable[[Optional[T], Optional[T]], Optional[T]]:
    def wrapper(opt1, opt2):
        if opt1 is None:
            return opt2
        if opt2 is None:
            return opt1
        return func(opt1, opt2)

    return wrapper


K = TypeVar("K")
V = TypeVar("V")


@merge_opts
def merge_opts_mapping(m1: Mapping[K, V], m2: Mapping[K, V]) -> Mapping[K, V]:
    return {**m1, **m2}


def is_type_var(cls: AnyType) -> bool:
    return isinstance(cls, TypeVar)  # type: ignore


def map_values(mapper: Callable[[V], T], mapping: Mapping[K, V]) -> Mapping[K, T]:
    return {k: mapper(v) for k, v in mapping.items()}


Func = TypeVar("Func", bound=Callable)


def typed_wraps(wrapped: Func) -> Callable[[Callable], Func]:
    return cast(Func, wraps(wrapped))


def get_origin_or_class(cls: AnyType) -> Type:
    origin = get_origin(cls)
    return origin if origin is not None else cls


try:
    from functools import cached_property
except ImportError:
    # From 3.9 functools
    from threading import RLock

    _NOT_FOUND = object()

    class cached_property:  # type: ignore
        def __init__(self, func):
            self.func = func
            self.attrname = None
            self.__doc__ = func.__doc__
            self.lock = RLock()

        def __set_name__(self, owner, name):
            if self.attrname is None:
                self.attrname = name
            elif name != self.attrname:
                raise TypeError(
                    "Cannot assign the same cached_property to two different names "
                    f"({self.attrname!r} and {name!r})."
                )

        def __get__(self, instance, owner=None):
            if instance is None:
                return self
            if self.attrname is None:
                raise TypeError(
                    "Cannot use cached_property instance"
                    " without calling __set_name__ on it."
                )
            try:
                cache = instance.__dict__
            except AttributeError:  # not all objects have __dict__ (e.g. class defines slots) # noqa: E501
                msg = (
                    f"No '__dict__' attribute on {type(instance).__name__!r} "
                    f"instance to cache {self.attrname!r} property."
                )
                raise TypeError(msg) from None
            val = cache.get(self.attrname, _NOT_FOUND)
            if val is _NOT_FOUND:
                with self.lock:
                    # check if another thread filled cache while we awaited lock
                    val = cache.get(self.attrname, _NOT_FOUND)
                    if val is _NOT_FOUND:
                        val = self.func(instance)
                        try:
                            cache[self.attrname] = val
                        except TypeError:
                            msg = (
                                f"The '__dict__' attribute on"
                                f" {type(instance).__name__!r} instance "
                                f"does not support item assignment for"
                                f" caching {self.attrname!r} property."
                            )
                            raise TypeError(msg) from None
            return val

        __class_getitem__ = classmethod(_GenericAlias)



import sys
from dataclasses import (  # type: ignore
    Field,
    InitVar,
    MISSING,
    _FIELDS,
    _FIELD_CLASSVAR,
    is_dataclass,
    make_dataclass,
)
from functools import lru_cache, partial
from types import MappingProxyType
from typing import (
    Any,
    Callable,
    Collection,
    Mapping,
    Sequence,
    Tuple,
    Type,
)

from apischema.dependent_required import (
    DEPENDENT_REQUIRED_ATTR,
    DependentRequired,
    Requirements,
)
from apischema.types import AnyType
from apischema.typing import get_origin, get_type_hints
from apischema.utils import PREFIX

if sys.version_info <= (3, 7):  # pragma: no cover
    is_dataclass_ = is_dataclass

    def is_dataclass(obj) -> bool:
        return is_dataclass_(obj) and getattr(obj, "__origin__", None) is None


@lru_cache()
def dataclass_types_and_fields(
    cls: Type,
) -> Tuple[Mapping[str, AnyType], Sequence[Field], Sequence[Field]]:
    from apischema.metadata.keys import INIT_VAR_METADATA

    assert is_dataclass(cls)
    types = get_type_hints(cls, include_extras=True)
    fields, init_fields = [], []
    for field in getattr(cls, _FIELDS).values():
        assert isinstance(field, Field)
        if field._field_type == _FIELD_CLASSVAR:  # type: ignore
            continue
        field_type = types[field.name]
        if isinstance(field_type, InitVar):
            types[field.name] = field_type.type  # type: ignore
            init_fields.append(field)
        elif field_type is InitVar:
            metadata = getattr(cls, _FIELDS)[field.name].metadata
            if INIT_VAR_METADATA not in metadata:
                raise TypeError("Before 3.8, InitVar requires init_var metadata")
            init_field = (PREFIX, metadata[INIT_VAR_METADATA], ...)
            tmp_cls = make_dataclass("Tmp", [init_field], bases=(cls,))  # type: ignore
            types[field.name] = get_type_hints(tmp_cls, include_extras=True)[PREFIX]
            init_fields.append(field)
        else:
            fields.append(field)
    # Use immutable return because of cache
    return MappingProxyType(types), tuple(fields), tuple(init_fields)


def has_default(field: Field) -> bool:
    return field.default is not MISSING or field.default_factory is not MISSING  # type: ignore # noqa: E501


def is_required(field: Field) -> bool:
    from apischema.metadata.keys import REQUIRED_METADATA

    return REQUIRED_METADATA in field.metadata or not has_default(field)


def get_default(field: Field) -> Any:
    if field.default_factory is not MISSING:  # type: ignore
        return field.default_factory()  # type: ignore
    if field.default is not MISSING:
        return field.default
    raise NotImplementedError()


def get_alias(field: Field) -> str:
    from apischema.metadata.keys import ALIAS_METADATA

    return field.metadata.get(ALIAS_METADATA, field.name)


def _merge_requirements(
    cls: Type,
    method: Callable[[DependentRequired], Requirements],
) -> Tuple[Requirements, Requirements]:
    assert is_dataclass(cls)
    _, _, init_vars = dataclass_types_and_fields(cls)  # type: ignore
    init_only = {f.name for f in init_vars}
    all_dependent_required: Collection["DependentRequired"] = getattr(
        cls, DEPENDENT_REQUIRED_ATTR, ()
    )
    deserialization_requirements = {}
    serialization_requirements = {}
    for dep_req in all_dependent_required:
        for field, required in method(dep_req).items():  # noqa: F402
            deserialization_requirements[field] = {req for req in required if req.init}
            serialization_requirements[field] = {
                req for req in required if req not in init_only
            }
    return deserialization_requirements, serialization_requirements


get_required_by = partial(_merge_requirements, method=DependentRequired.required_by)
get_requiring = partial(_merge_requirements, method=DependentRequired.requiring)


def check_merged_class(merged_cls: AnyType) -> Type:
    origin = get_origin(merged_cls)
    if origin is None:
        origin = merged_cls
    if not is_dataclass(origin):
        raise TypeError("Merged field must have dataclass type")
    return origin



from dataclasses import Field
from typing import Iterable, Mapping, Optional, Sequence, Tuple, Type

from apischema.conversions.metadata import get_field_conversions
from apischema.conversions.utils import Conversions
from apischema.conversions.visitor import (
    Conv,
    ConversionsVisitor,
    Deserialization,
    DeserializationVisitor,
    Serialization,
    SerializationVisitor,
)
from apischema.dataclass_utils import (
    Requirements,
    get_requiring,
)
from apischema.metadata.keys import SKIP_METADATA
from apischema.resolvers import Resolver, get_resolvers
from apischema.skip import filter_skipped
from apischema.types import AnyType
from apischema.visitor import Return


class SchemaVisitor(ConversionsVisitor[Conv, Return]):
    def _dataclass_fields_(
        self, fields: Sequence[Field], init_vars: Sequence[Field]
    ) -> Iterable[Field]:
        raise NotImplementedError()

    def _dataclass_fields(
        self, fields: Sequence[Field], init_vars: Sequence[Field]
    ) -> Sequence[Field]:
        return [
            f
            for f in self._dataclass_fields_(fields, init_vars)
            if SKIP_METADATA not in f.metadata
        ]

    def _field_conversions(
        self, field: Field, field_type: AnyType
    ) -> Tuple[AnyType, Optional[Conversions]]:
        raise NotImplementedError()

    def _visit_field_(self, field: Field, field_type: AnyType) -> Return:
        return self.visit(field_type)

    def _visit_field(self, field: Field, field_type: AnyType) -> Return:
        field_type, conversions = self._field_conversions(field, field_type)
        with self._replace_conversions(conversions):
            return self._visit_field_(field, field_type)

    def _dependent_required(self, cls: Type) -> Requirements:
        raise NotImplementedError()

    def _resolvers(self, cls: Type) -> Mapping[str, Resolver]:
        raise NotImplementedError()

    def union(self, alternatives: Sequence[AnyType]) -> Return:
        return self._union_result(
            map(self.visit, filter_skipped(alternatives, schema_only=True))
        )


class DeserializationSchemaVisitor(
    DeserializationVisitor[Return], SchemaVisitor[Deserialization, Return]
):
    def visit_conversion(self, cls: AnyType, conversion: Deserialization) -> Return:
        return self._union_result(
            [
                self.visit_with_conversions(source, conversions)
                for source, (_, conversions) in conversion.items()
            ]
        )

    @staticmethod
    def _dataclass_fields_(
        fields: Sequence[Field], init_vars: Sequence[Field]
    ) -> Sequence[Field]:
        return (*(f for f in fields if f.init), *init_vars)

    @staticmethod
    def _field_conversions(
        field: Field, field_type: AnyType
    ) -> Tuple[AnyType, Optional[Conversions]]:
        conversions = get_field_conversions(field, field_type)
        if conversions is None:
            return field_type, None
        elif conversions.deserializer is None:
            return field_type, conversions.deserialization
        else:
            type_ = next(iter(conversions.deserialization_conversion(field_type)))
            return type_, conversions.deserialization

    def _dependent_required(self, cls: Type) -> Requirements:
        return get_requiring(cls)[0]

    def _resolvers(self, cls: Type) -> Mapping[str, Resolver]:
        return {}


class SerializationSchemaVisitor(
    SerializationVisitor[Return], SchemaVisitor[Serialization, Return]
):
    def visit_conversion(self, cls: AnyType, conversion: Serialization) -> Return:
        target, (_, conversions) = conversion
        return self.visit_with_conversions(target, conversions)

    @staticmethod
    def _dataclass_fields_(
        fields: Sequence[Field], init_vars: Sequence[Field]
    ) -> Sequence[Field]:
        return fields

    @staticmethod
    def _field_conversions(
        field: Field, field_type: AnyType
    ) -> Tuple[AnyType, Optional[Conversions]]:
        conversions = get_field_conversions(field, field_type)
        if conversions is None:
            return field_type, None
        elif conversions.serializer is None:
            return field_type, conversions.serialization
        else:
            type_, *_ = conversions.serialization_conversion(field_type)
            return type_, conversions.serialization

    @staticmethod
    def _dependent_required(cls: Type) -> Requirements:
        return get_requiring(cls)[1]

    def _resolvers(self, cls: Type) -> Mapping[str, Resolver]:
        return get_resolvers(cls)



from collections import ChainMap
from dataclasses import Field, dataclass, fields
from types import MappingProxyType
from typing import Callable, Mapping, Optional, Type, TypeVar, overload

from apischema.conversions.raw import to_raw_deserializer
from apischema.conversions.utils import (
    Conversions,
    Converter,
    check_converter,
)
from apischema.conversions.visitor import Deserialization, Serialization
from apischema.dataclass_utils import is_dataclass
from apischema.type_vars import resolve_type_vars
from apischema.types import AnyType, Metadata, MetadataMixin
from apischema.typing import get_args, get_origin
from apischema.utils import is_type_var


def handle_generic_field_type(
    field_type: AnyType, base: AnyType, other: AnyType, covariant: bool
) -> AnyType:
    contravariant = not covariant
    type_vars = None
    if is_type_var(base):
        type_vars = {base: field_type}
    if (
        get_origin(base) is not None
        and getattr(base, "__parameters__", ())
        and len(get_args(base)) == len(get_args(field_type))
        and not any(map(get_origin, get_args(base)))
        and not any(map(get_origin, get_args(field_type)))
        and not any(
            not is_type_var(base_arg) and base_arg != field_arg
            for base_arg, field_arg in zip(get_args(base), get_args(field_type))
        )
    ):
        type_vars = {}
        for base_arg, field_arg in zip(get_args(base), get_args(field_type)):
            if base_arg in type_vars and type_vars[base_arg] != field_arg:
                type_vars = None
                break
            type_vars[base_arg] = field_arg
        field_type_origin, base_origin = get_origin(field_type), get_origin(base)
        assert field_type_origin is not None and base_origin is not None
        if base_origin != field_type_origin:
            if covariant and not issubclass(base_origin, field_type_origin):
                type_vars = None
            if contravariant and not issubclass(field_type_origin, base_origin):
                type_vars = None
    return resolve_type_vars(other, type_vars)


Cls = TypeVar("Cls", bound=Type)


@dataclass
class FieldConversions(MetadataMixin):
    deserialization: Optional[Conversions] = None
    serialization: Optional[Conversions] = None
    deserializer: Optional[Converter] = None
    serializer: Optional[Converter] = None

    def __post_init__(self):
        from apischema.metadata.keys import CONVERSIONS_METADATA

        super().__init__(CONVERSIONS_METADATA)

    def __call__(self, dataclass: Cls) -> Cls:
        from apischema.metadata.keys import CONVERSIONS_METADATA, MERGED_METADATA

        if not is_dataclass(dataclass):
            raise TypeError("Must be applied to dataclass")
        for conv in (self.deserialization, self.serialization):
            if conv is not None and not isinstance(conv, Mapping):
                raise TypeError("Dataclasses conversions must be a Mapping instance")
        for field in fields(dataclass):
            if field.metadata.get(MERGED_METADATA):
                continue
            if CONVERSIONS_METADATA in field.metadata:
                conversions: FieldConversions = field.metadata[CONVERSIONS_METADATA]
                if any((self.deserialization, conversions.deserialization)):
                    conversions.deserialization = ChainMap(
                        conversions.deserialization or {}, self.deserialization or {}
                    )
                if any((self.serialization, conversions.serialization)):
                    conversions.serialization = ChainMap(
                        conversions.serialization or {}, self.serialization or {}
                    )
            else:
                conversions = FieldConversions(
                    deserialization=self.deserialization,
                    serialization=self.serialization,
                )
                field.metadata = MappingProxyType({CONVERSIONS_METADATA: conversions})
        return dataclass

    def deserialization_conversion(self, field_type: AnyType) -> Deserialization:
        assert self.deserializer is not None
        try:
            param, ret = check_converter(self.deserializer, None, None)
        except TypeError:
            param, _ = check_converter(self.deserializer, None, field_type)
        else:
            param = handle_generic_field_type(field_type, ret, param, True)
        return {param: (self.deserializer, self.deserialization)}

    def serialization_conversion(self, field_type: AnyType) -> Serialization:
        assert self.serializer is not None
        try:
            param, ret = check_converter(self.serializer, None, None)
        except TypeError:
            _, ret = check_converter(self.serializer, field_type, None)
        else:
            ret = handle_generic_field_type(field_type, param, ret, False)
        return ret, (self.serializer, self.serialization)


@dataclass
class FieldConversionsModel(MetadataMixin):
    model: AnyType

    def __post_init__(self):
        from apischema.metadata.keys import CONVERSIONS_METADATA

        super().__init__(CONVERSIONS_METADATA)


@overload
def conversions(model: AnyType) -> Metadata:
    ...


@overload
def conversions(
    *,
    deserialization: Conversions = None,
    serialization: Conversions = None,
    deserializer: Converter = None,
    serializer: Converter = None,
    raw_deserializer: Callable = None,
) -> FieldConversions:
    ...


def conversions(
    model: AnyType = None,
    *,
    deserialization: Conversions = None,
    serialization: Conversions = None,
    deserializer: Converter = None,
    serializer: Converter = None,
    raw_deserializer: Callable = None,
) -> Metadata:
    if model is not None:
        return FieldConversionsModel(model)
    else:
        if raw_deserializer is not None:
            deserializer = to_raw_deserializer(raw_deserializer)
        return FieldConversions(
            deserialization, serialization, deserializer, serializer
        )


def get_field_conversions(
    field: Field, field_type: AnyType
) -> Optional[FieldConversions]:
    from apischema.metadata.keys import CONVERSIONS_METADATA

    if CONVERSIONS_METADATA not in field.metadata:
        return None
    else:
        conversions = field.metadata[CONVERSIONS_METADATA]
        if isinstance(conversions, FieldConversionsModel):
            return FieldConversions(
                {field_type: conversions.model}, {field_type: conversions.model}
            )
        else:
            return conversions



import warnings
from contextlib import contextmanager
from typing import Any, Generic, Mapping, Optional, Sequence, Tuple, Type, TypeVar

from apischema.conversions.converters import (
    _deserializers,
    _extra_deserializers,
    _extra_serializers,
    _serializers,
)
from apischema.conversions.dataclass_model import DataclassModelWrapper, get_model
from apischema.conversions.utils import (
    Conversions,
    ConverterWithConversions,
    get_parameters,
)
from apischema.type_vars import resolve_type_vars
from apischema.types import AnyType
from apischema.visitor import Return, Visitor

Conv = TypeVar("Conv")
Deserialization = Mapping[Type, ConverterWithConversions]
Serialization = Tuple[Type, ConverterWithConversions]


class ConversionsVisitor(Generic[Conv, Return], Visitor[Return]):
    def __init__(self):
        super().__init__()
        self._conversions = None

    def _is_conversion(self, cls: Type, arg: Optional[Any]) -> Optional[Conv]:
        raise NotImplementedError()

    def is_conversion(self, cls: Type) -> Optional[Conv]:
        arg = None
        if self._conversions is not None:
            try:
                arg = self._conversions[cls]
            except KeyError:
                pass
        return self._is_conversion(cls, arg)

    def is_extra_conversions(self, cls: AnyType) -> bool:
        return (
            isinstance(cls, type)
            and self._conversions is not None
            and cls in self._conversions
            and self.is_conversion(cls) is not None
        )

    def visit_conversion(self, cls: AnyType, conversion: Conv) -> Return:
        raise NotImplementedError()

    def visit_not_conversion(self, cls: AnyType) -> Return:
        return super()._visit(cls)

    @contextmanager
    def _replace_conversions(self, conversions: Optional[Conversions]):
        conversions_save = self._conversions
        self._conversions = conversions
        try:
            yield
        finally:
            self._conversions = conversions_save

    def _visit(self, cls: Type) -> Return:
        if not isinstance(cls, type):
            if isinstance(cls, DataclassModelWrapper):
                return self.visit(get_model(cls.cls, cls.model))
            else:
                return self.visit_not_conversion(cls)
        conversion = self.is_conversion(cls)
        with self._replace_conversions(None):
            if conversion is not None:
                return self.visit_conversion(cls, conversion)
            else:
                return self.visit_not_conversion(cls)

    def visit_with_conversions(
        self, cls: AnyType, conversions: Optional[Conversions]
    ) -> Return:
        with self._replace_conversions(conversions):
            return self.visit(cls)


def handle_generic_conversion(base: AnyType, other: AnyType) -> AnyType:
    if isinstance(other, tuple):
        type_vars, other = other
        return resolve_type_vars(other, dict(zip(type_vars, get_parameters(base))))
    else:
        return other


class DeserializationVisitor(ConversionsVisitor[Deserialization, Return]):
    @staticmethod
    def _is_conversion(cls: Type, source: Optional[Any]) -> Optional[Deserialization]:
        if source is cls:
            return None
        if source is not None and cls in _extra_deserializers:
            sources = source if isinstance(source, Sequence) else [source]
            result = {}
            for source in sources:
                source2 = handle_generic_conversion(cls, source)
                if source2 not in _extra_deserializers[cls]:
                    warnings.warn(f"Deserializer {source} -> {cls} doesn't exists")
                else:
                    result[source2] = _extra_deserializers[cls][source2]
            return result
        return _deserializers.get(cls)


class SerializationVisitor(ConversionsVisitor[Serialization, Return]):
    @staticmethod
    def _is_conversion(cls: Type, target: Optional[Any]) -> Optional[Serialization]:
        if target is cls:
            return None
        if target is not None:
            for sub_cls in cls.__mro__:
                if sub_cls in _extra_serializers:
                    target2 = handle_generic_conversion(sub_cls, target)
                    if target2 not in _extra_serializers[sub_cls]:
                        warnings.warn(f"Serializer {cls} -> {target} doesn't exists")
                    else:
                        return target2, _extra_serializers[sub_cls][target2]
        for sub_cls in cls.__mro__:
            if sub_cls in _serializers:
                return _serializers[sub_cls]
        else:
            return None



from typing import Any, Iterable, Mapping, Optional, TypeVar, Union

from apischema.types import AnyType
from apischema.typing import get_args, get_origin
from apischema.utils import is_type_var

TV = AnyType  # TypeVar is not supported as a type

# 10 should be enough for all builtin types
# but it should be an infinite list with locked appending
_type_vars = [TypeVar(f"T{i}") for i in range(10)]


def get_parameters(cls: AnyType) -> Iterable[TV]:
    return getattr(cls, "__parameters__", _type_vars)


TypeVarContext = Optional[Mapping[TV, AnyType]]


def resolve_type_vars(cls: AnyType, type_vars: TypeVarContext = None) -> Any:
    if is_type_var(cls):
        if type_vars is not None and cls in type_vars:
            return type_vars[cls]
        elif cls.__constraints__:
            return Union[cls.__constraints__]
        else:
            return Any
    elif getattr(cls, "__parameters__", ()):
        return cls[tuple(resolve_type_vars(p, type_vars) for p in cls.__parameters__)]
    else:
        return cls


def type_var_context(cls: AnyType, type_vars: TypeVarContext = None) -> TypeVarContext:
    cls = resolve_type_vars(cls, type_vars)
    origin = get_origin(cls)
    assert origin is not None
    return dict(zip(get_parameters(origin), get_args(cls)))



from inspect import Parameter, isclass, signature
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    cast,
)

from apischema.type_vars import get_parameters, resolve_type_vars
from apischema.types import AnyType
from apischema.typing import get_args, get_origin, get_type_hints
from apischema.utils import is_type_var, type_name
from apischema.visitor import Unsupported, Visitor

Conversions = Mapping[AnyType, Any]
Converter = Callable[[Any], Any]
ConverterWithConversions = Tuple[Converter, Optional[Conversions]]


def check_converter(
    converter: Converter,
    param: Optional[AnyType],
    ret: Optional[AnyType],
    namespace: Dict[str, Any] = None,
) -> Tuple[AnyType, AnyType]:
    try:
        parameters = iter(signature(converter).parameters.values())
    except ValueError:  # builtin types
        if ret is None and isclass(converter):
            ret = cast(Type[Any], converter)
        if param is None:
            raise TypeError("converter parameter must be typed")
    else:
        try:
            first = next(parameters)
        except StopIteration:
            raise TypeError("converter must have at least one parameter")
        types = get_type_hints(converter, None, namespace, include_extras=True)
        for p in parameters:
            if p.default is Parameter.empty and p.kind not in (
                Parameter.VAR_POSITIONAL,
                Parameter.VAR_KEYWORD,
            ):
                raise TypeError(
                    "converter must have at most one parameter " "without default"
                )
        if param is None:
            try:
                param = types.pop(first.name)
            except KeyError:
                raise TypeError("converter parameter must be typed")
        if ret is None:
            try:
                ret = types.pop("return")
            except KeyError:
                if isclass(converter):
                    ret = cast(Type, converter)
                else:
                    raise TypeError("converter return must be typed")
    return param, ret


def handle_generic_conversions(
    base: AnyType, other: AnyType
) -> Tuple[AnyType, AnyType]:
    origin = get_origin(base)
    if origin is None:
        return base, other
    args = get_args(base)
    if not all(map(is_type_var, args)):
        raise TypeError(
            f"Generic conversion doesn't support specialization,"
            f" aka {type_name(base)}[{','.join(map(type_name, args))}]"
        )
    return origin, resolve_type_vars(other, dict(zip(args, get_parameters(origin))))


class ConvertibleVisitor(Visitor[bool]):
    def annotated(self, cls: AnyType, annotations: Sequence[Any]) -> bool:
        return False

    def any(self) -> bool:
        return False

    def collection(self, cls: Type[Collection], value_type: AnyType) -> bool:
        return False

    def generic(self, cls: AnyType) -> bool:
        return True

    def literal(self, values: Sequence[Any]) -> bool:
        return False

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> bool:
        return False

    def new_type(self, cls: AnyType, super_type: AnyType) -> bool:
        return False

    def tuple(self, types: Sequence[AnyType]) -> bool:
        return False

    def union(self, alternatives: Sequence[AnyType]) -> bool:
        return False


def is_convertible(cls: AnyType):
    try:
        return ConvertibleVisitor().visit(cls)
    except (NotImplementedError, Unsupported):
        return True


def check_convertible(cls: AnyType):
    if not is_convertible(cls):
        raise TypeError(f"{type_name(cls)} is not a class")


class ConversionsWrapper:
    """Allows to hash conversions — conversions must be immutable"""

    def __init__(self, conversions: Conversions):
        self.conversions = conversions

    def __hash__(self):
        return id(self.conversions)


identity = lambda x: x  # noqa: E731



from collections import ChainMap
from contextlib import suppress
from dataclasses import Field, InitVar, dataclass
from enum import Enum
from inspect import Parameter
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Awaitable,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import graphql

from apischema import serialize
from apischema.aliases import Aliaser
from apischema.conversions import Conversions, Deserialization, Serialization
from apischema.conversions.visitor import Conv
from apischema.dataclass_utils import (
    check_merged_class,
    get_alias,
    get_default,
    is_required,
)
from apischema.graphql.interfaces import get_interfaces, is_interface
from apischema.graphql.resolvers import INFO_TYPES, resolver_resolve
from apischema.json_schema.generation.visitor import (
    DeserializationSchemaVisitor,
    SchemaVisitor,
    SerializationSchemaVisitor,
)
from apischema.json_schema.refs import get_ref, schema_ref
from apischema.json_schema.schema import Schema, get_schema, merge_schema
from apischema.metadata.keys import (
    MERGED_METADATA,
    PROPERTIES_METADATA,
    SCHEMA_METADATA,
    check_metadata,
)
from apischema.resolvers import (
    MissingFirstParameter,
    Resolver,
    get_resolvers,
    resolver_parameters,
)
from apischema.skip import filter_skipped
from apischema.types import AnyType, NoneType
from apischema.typing import get_args, get_origin
from apischema.utils import Undefined, is_hashable, map_values, to_camel_case, type_name

JsonScalar = graphql.GraphQLScalarType(
    "JSON",
    specified_by_url="http://www.ecma-international.org/publications/files/ECMA-ST/ECMA-404.pdf",  # noqa: E501
)
GRAPHQL_PRIMITIVE_TYPES = {
    int: graphql.GraphQLInt,
    float: graphql.GraphQLFloat,
    str: graphql.GraphQLString,
    bool: graphql.GraphQLBoolean,
}


class MissingRef(Exception):
    pass


class Nullable(Exception):
    pass


def ref_or_name(cls: AnyType) -> str:
    return get_ref(cls) or type_name(cls)


T = TypeVar("T")
Lazy = Callable[[], T]
Thunk = Union[Callable[[], T], T]


def exec_thunk(thunk: Thunk[graphql.GraphQLType], *, non_null=None) -> Any:
    result = thunk if isinstance(thunk, graphql.GraphQLType) else thunk()
    if non_null is True and not isinstance(result, graphql.GraphQLNonNull):
        return graphql.GraphQLNonNull(result)  # type: ignore
    if non_null is False and isinstance(result, graphql.GraphQLNonNull):
        return result.of_type
    return result


U = TypeVar("U")


@dataclass(frozen=True)
class FieldParameter:
    name: str
    type: AnyType
    default: Any


def field_parameters(resolver: Resolver) -> Sequence[FieldParameter]:
    return [
        FieldParameter(p.name, resolver.types[p.name], p.default)
        for p in resolver.parameters
        if resolver.types[p.name] not in INFO_TYPES
    ]


@dataclass(frozen=True)
class ObjectField:
    name: str
    type: AnyType
    alias: Optional[str] = None
    conversions: Optional[Conversions] = None
    default: Any = graphql.Undefined
    deprecated: Optional[str] = None
    description: Optional[str] = None
    parameters: Optional[Collection[FieldParameter]] = None
    required: InitVar[bool] = True
    resolve: Optional[Callable] = None
    schema: InitVar[Optional[Schema]] = None
    subscribe: Optional[Callable] = None

    def __post_init__(self, required: bool, schema: Optional[Schema]):
        if required:
            object.__setattr__(self, "default", graphql.Undefined)
        if schema is not None and schema.annotations is not None:
            object.__setattr__(self, "description", schema.annotations.description)
            if schema.annotations.deprecated is True:
                object.__setattr__(
                    self, "deprecated", graphql.DEFAULT_DEPRECATION_REASON
                )
            elif isinstance(schema.annotations.deprecated, str):
                object.__setattr__(self, "deprecated", schema.annotations.deprecated)
            if schema.annotations.default is not Undefined and not required:
                object.__setattr__(self, "default", schema.annotations.default)

    def apply_aliaser(self, aliaser: Aliaser) -> str:
        return aliaser(self.alias or self.name)


def wrap_return_type(return_type: AnyType, error_as_null: bool) -> AnyType:
    return Optional[return_type] if error_as_null else return_type


IdPredicate = Callable[[AnyType], bool]
GenericRefFactory = Callable[[AnyType], str]
UnionRefFactory = Callable[[Sequence[str]], str]


class SchemaBuilder(SchemaVisitor[Conv, Thunk[graphql.GraphQLType]]):
    def __init__(
        self,
        aliaser: Optional[Aliaser],
        is_id: Optional[IdPredicate],
        error_as_null: bool,
        generic_ref_factory: Optional[GenericRefFactory],
        union_ref_factory: Optional[UnionRefFactory],
    ):
        super().__init__()
        self.aliaser = aliaser or (lambda s: s)
        self.is_id = is_id or (lambda t: False)
        self.error_as_null = error_as_null
        self.generic_ref_factory = generic_ref_factory
        self.union_ref_factory = union_ref_factory
        self._cache: Dict[Any, Thunk[graphql.GraphQLType]] = {}
        self._non_null = True
        self._ref: Optional[str] = None
        self._schema: Optional[Schema] = None

    @property
    def _description(self) -> Optional[str]:
        if self._schema is not None and self._schema.annotations is not None:
            return self._schema.annotations.description
        else:
            return None

    @property
    def _ref_and_desc(self) -> Tuple[str, Optional[str]]:
        if self._ref is None:
            raise MissingRef
        return self._ref, self._description

    def annotated(
        self, cls: AnyType, annotations: Sequence[Any]
    ) -> Thunk[graphql.GraphQLType]:
        for annotation in annotations:
            if isinstance(annotation, schema_ref):
                annotation.check_type(cls)
                ref = annotation.ref
                if not isinstance(ref, str):
                    raise ValueError("Annotated schema_ref can only be str")
                self._ref = self._ref or ref
            if isinstance(annotation, Schema):
                self._schema = merge_schema(annotation, self._schema)
        return self.visit_with_schema(cls, self._ref, self._schema)

    def any(self) -> Thunk[graphql.GraphQLType]:
        return JsonScalar

    def collection(
        self, cls: Type[Collection], value_type: AnyType
    ) -> Thunk[graphql.GraphQLType]:
        value_thunk = self.visit(value_type)
        return lambda: graphql.GraphQLList(exec_thunk(value_thunk))

    def _object_field(self, field: Field, field_type: AnyType) -> ObjectField:
        field_type, conversions = self._field_conversions(field, field_type)
        default: Any = graphql.Undefined
        if not is_required(field):
            with suppress(Exception):
                default = serialize(get_default(field), conversions=conversions)
        return ObjectField(
            field.name,
            field_type,
            alias=get_alias(field),
            conversions=conversions,
            default=default,
            required=is_required(field),
            schema=field.metadata.get(SCHEMA_METADATA),
        )

    def dataclass(
        self,
        cls: Type,
        types: Mapping[str, AnyType],
        fields: Sequence[Field],
        init_vars: Sequence[Field],
    ) -> Thunk[graphql.GraphQLType]:
        types = dict(types)
        object_fields: List[ObjectField] = []
        merged_fields: Dict[str, AnyType] = {}
        for field in self._dataclass_fields(fields, init_vars):
            check_metadata(field)
            metadata = field.metadata
            if MERGED_METADATA in metadata:
                check_merged_class(types[field.name])
                merged_fields[field.name] = types[field.name]
            elif PROPERTIES_METADATA in metadata:
                continue
            else:
                object_fields.append(self._object_field(field, types[field.name]))
        return self.object(cls, object_fields, merged_fields)

    def enum(self, cls: Type[Enum]) -> Thunk[graphql.GraphQLType]:
        return self.literal([elt.value for elt in cls])

    def generic(self, cls: AnyType) -> Thunk[graphql.GraphQLType]:
        if self.generic_ref_factory is None:
            raise MissingRef
        self._ref = self.generic_ref_factory(cls)
        return super().generic(cls)

    def literal(self, values: Sequence[Any]) -> Thunk[graphql.GraphQLType]:
        if not all(isinstance(v, str) for v in values):
            raise TypeError("Apischema GraphQL only support Enum/Literal of strings")
        name, description = self._ref_and_desc
        return graphql.GraphQLEnumType(
            name, dict(zip(values, values)), description=description
        )

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> Thunk[graphql.GraphQLType]:
        try:
            name, description = self._ref_and_desc
            return graphql.GraphQLScalarType(name, description=description)
        except MissingRef:
            return JsonScalar

    def named_tuple(
        self,
        cls: Type[Tuple],
        types: Mapping[str, AnyType],
        defaults: Mapping[str, Any],
    ) -> Thunk[graphql.GraphQLType]:
        fields = []
        for field_name, field_type in types.items():
            default = graphql.Undefined
            if field_name in defaults:
                with suppress(Exception):
                    default = serialize(defaults[field_name])
            fields.append(ObjectField(field_name, field_type, default=default))
        return self.object(cls, fields)

    def new_type(self, cls: Type, super_type: AnyType) -> Thunk[graphql.GraphQLType]:
        return self.visit_with_schema(super_type, self._ref, self._schema)

    def object(
        self,
        cls: Type,
        fields: Collection[ObjectField],
        merged_fields: Mapping[str, Thunk[graphql.GraphQLType]] = None,
    ) -> Thunk[graphql.GraphQLType]:
        raise NotImplementedError()

    def primitive(self, cls: Type) -> Thunk[graphql.GraphQLType]:
        if cls is NoneType:
            raise Nullable
        try:
            name, description = self._ref_and_desc
            return graphql.GraphQLScalarType(name, description=description)
        except MissingRef:
            return GRAPHQL_PRIMITIVE_TYPES[cls]

    def tuple(self, types: Sequence[AnyType]) -> Thunk[graphql.GraphQLType]:
        raise TypeError("Tuple are not supported")

    def typed_dict(
        self, cls: Type, keys: Mapping[str, AnyType], total: bool
    ) -> Thunk[graphql.GraphQLType]:
        raise TypeError("TyedDict are not supported")

    def union(self, alternatives: Sequence[AnyType]) -> Thunk[graphql.GraphQLType]:
        alternatives = list(filter_skipped(alternatives, schema_only=True))
        results = []
        filtered_alt = []
        for alt in alternatives:
            try:
                results.append(self.visit(alt))
                filtered_alt.append(alt)
            except Nullable:
                self._non_null = False
        if not results:
            raise TypeError("Empty union")
        return self._union_result(results)

    def visit_with_schema(
        self, cls: AnyType, ref: Optional[str], schema: Optional[Schema]
    ) -> Thunk[graphql.GraphQLType]:
        if self.is_id(cls):
            return graphql.GraphQLNonNull(graphql.GraphQLID)
        if is_hashable(cls) and not self.is_extra_conversions(cls):
            ref, schema = ref or get_ref(cls), merge_schema(get_schema(cls), schema)
        else:
            schema, ref = None, None
        ref_save, schema_save, non_null_save = self._ref, self._schema, self._non_null
        self._ref, self._schema, self._non_null = ref, schema, True
        try:
            result = super().visit(cls)
            non_null = self._non_null
            return lambda: exec_thunk(result, non_null=non_null)
        except MissingRef:
            raise TypeError(f"Missing ref for type {cls}")
        finally:
            self._ref, self._schema = ref_save, schema_save
            self._non_null = non_null_save

    def visit_not_conversion(self, cls: AnyType) -> Thunk[graphql.GraphQLType]:
        key = self._resolve_type_vars(cls), self._ref, self._schema
        if key in self._cache:
            return self._cache[key]
        cache = None

        def rec_sentinel() -> graphql.GraphQLType:
            assert cache is not None
            return cache

        self._cache[key] = rec_sentinel
        try:
            cache = exec_thunk(super().visit_not_conversion(cls))
        except Exception:
            del self._cache[key]
            raise
        else:
            return cache
        # if key not in self._cache:
        #     thunk = super().visit_not_conversion(cls)
        #     if isinstance(thunk, graphql.GraphQLType):
        #         self._cache[key] = thunk
        #     else:
        #         self._cache[key] = lru_cache(maxsize=1)(thunk)
        # return self._cache[key]

    def visit(self, cls: AnyType) -> Thunk[graphql.GraphQLType]:
        return self.visit_with_schema(cls, None, None)


def deref_merged_field(
    merged_attr: str, field: graphql.GraphQLField
) -> graphql.GraphQLField:
    def resolve(obj, info, **kwargs):
        return field.resolve(getattr(obj, merged_attr), info, **kwargs)

    return graphql.GraphQLField(**ChainMap(dict(resolve=resolve), field.to_kwargs()))


FieldType = TypeVar("FieldType", graphql.GraphQLInputField, graphql.GraphQLField)


def merge_fields(
    cls: Type,
    fields: Mapping[str, Lazy[FieldType]],
    merged_types: Mapping[str, Thunk[graphql.GraphQLType]],
    merged_field_modifier: Callable[[str, FieldType], FieldType] = None,
) -> Dict[str, FieldType]:
    all_merged_fields: Dict[str, FieldType] = {}
    for merged_name, merged_thunk in merged_types.items():
        merged_type = exec_thunk(merged_thunk, non_null=False)
        merged_fields: Mapping[str, FieldType] = merged_type.fields
        if merged_fields.keys() & all_merged_fields.keys() & fields.keys():
            raise TypeError(f"Conflict in merged fields of {cls}")
        if merged_field_modifier:
            merged_fields = {
                name: merged_field_modifier(merged_name, field)
                for name, field in merged_fields.items()
            }
        all_merged_fields.update(merged_fields)
    return {**{name: field() for name, field in fields.items()}, **all_merged_fields}


class InputSchemaBuilder(
    DeserializationSchemaVisitor[Thunk[graphql.GraphQLType]],
    SchemaBuilder[Deserialization],
):
    def _field(self, field: ObjectField) -> Tuple[str, Lazy[graphql.GraphQLInputField]]:
        field_type = self.visit_with_conversions(field.type, field.conversions)
        return field.apply_aliaser(self.aliaser), lambda: graphql.GraphQLInputField(
            exec_thunk(field_type),
            default_value=field.default,
            description=field.description,
            out_name=field.name,
        )

    def object(
        self,
        cls: Type,
        fields: Collection[ObjectField],
        merged_fields: Mapping[str, AnyType] = None,
    ) -> Thunk[graphql.GraphQLType]:
        name, description = self._ref_and_desc
        name = name if name.endswith("Input") else name + "Input"
        visited_fields = dict(map(self._field, fields))
        visited_merged = map_values(self.visit, merged_fields or {})
        return lambda: graphql.GraphQLInputObjectType(
            name, lambda: merge_fields(cls, visited_fields, visited_merged), description
        )

    def _union_result(
        self, results: Iterable[Thunk[graphql.GraphQLType]]
    ) -> Thunk[graphql.GraphQLType]:
        results = list(results)  # Execute the iteration
        if len(results) == 1:
            return results[0]
        raise TypeError("Union are not supported for input")


class OutputSchemaBuilder(
    SerializationSchemaVisitor[Thunk[graphql.GraphQLType]], SchemaBuilder[Serialization]
):
    def __init__(
        self,
        aliaser: Optional[Aliaser],
        is_id: Optional[IdPredicate],
        error_as_null: bool,
        generic_ref_factory: Optional[GenericRefFactory],
        union_ref_factory: Optional[UnionRefFactory],
    ):
        super().__init__(
            aliaser, is_id, error_as_null, generic_ref_factory, union_ref_factory
        )
        self.input_builder = InputSchemaBuilder(
            aliaser, is_id, error_as_null, generic_ref_factory, union_ref_factory
        )

    def _field(self, field: ObjectField) -> Tuple[str, Lazy[graphql.GraphQLField]]:
        alias = field.apply_aliaser(self.aliaser)
        if field.resolve is not None:
            resolve = field.resolve
        else:
            resolve = lambda obj, _: getattr(obj, field.name)  # noqa: E731
        field_type = self.visit_with_conversions(field.type, field.conversions)
        args = None
        if field.parameters is not None:
            args = {}
            for param in field.parameters:
                default = graphql.Undefined
                if param.default not in {Parameter.empty, Undefined, graphql.Undefined}:
                    with suppress(Exception):
                        default = serialize(param.default)

                def arg_thunk(
                    arg_thunk=self.input_builder.visit(param.type),
                    default=default,
                    out_name=param.name,
                ) -> graphql.GraphQLArgument:
                    arg_type = exec_thunk(arg_thunk)
                    if (
                        not isinstance(arg_type, graphql.GraphQLNonNull)
                        and default is None
                    ):
                        default = graphql.Undefined
                    return graphql.GraphQLArgument(arg_type, default, out_name=out_name)

                args[self.aliaser(param.name)] = arg_thunk
        return alias, lambda: graphql.GraphQLField(
            exec_thunk(field_type),
            {name: arg() for name, arg in args.items()} if args else None,
            resolve,
            field.subscribe,
            field.description,
            field.deprecated,
        )

    def object(
        self,
        cls: Type,
        fields: Collection[ObjectField],
        merged_fields: Mapping[str, AnyType] = None,
    ) -> Thunk[graphql.GraphQLType]:
        fields_and_resolvers = list(fields)
        try:
            name, description = self._ref_and_desc
        except MissingRef:
            if cls.__name__ not in ("Query", "Mutation", "Subscription"):
                raise
            name, description = cls.__name__, self._description
        for resolver_name, resolver in get_resolvers(cls).items():
            resolve = resolver_resolve(
                resolver, self.aliaser, error_as_null=self.error_as_null
            )
            resolver_field = ObjectField(
                resolver_name,
                wrap_return_type(resolver.return_type, self.error_as_null),
                conversions=resolver.conversions,
                parameters=field_parameters(resolver),
                resolve=resolve,
            )
            fields_and_resolvers.append(resolver_field)
        visited_fields = dict(map(self._field, fields_and_resolvers))
        visited_merged = map_values(self.visit, merged_fields or {})

        def field_thunk() -> graphql.GraphQLFieldMap:
            return merge_fields(
                cls,
                visited_fields,
                visited_merged,
                deref_merged_field,
            )

        interfaces = list(map(self.visit, get_interfaces(cls)))
        interface_thunk = None
        if interfaces:

            def interface_thunk() -> Collection[graphql.GraphQLInterfaceType]:
                result = {exec_thunk(i, non_null=False) for i in interfaces}
                for merged_thunk in visited_merged.values():
                    merged = cast(
                        Union[graphql.GraphQLObjectType, graphql.GraphQLInterfaceType],
                        exec_thunk(merged_thunk, non_null=False),
                    )
                    result.update(merged.interfaces)
                return sorted(result, key=lambda i: i.name)

        if is_interface(cls):
            return lambda: graphql.GraphQLInterfaceType(
                name,
                field_thunk,
                interface_thunk,
                description=description,
            )

        else:
            return lambda: graphql.GraphQLObjectType(
                name,
                field_thunk,
                interface_thunk,
                is_type_of=lambda obj, _: isinstance(obj, cls),
                description=description,
            )

    def _union_result(
        self, results: Iterable[Thunk[graphql.GraphQLType]]
    ) -> Thunk[graphql.GraphQLType]:
        results = list(results)  # Execute the iteration
        if len(results) == 1:
            return results[0]
        name, description = self._ref, self._description
        if name is None and self.union_ref_factory is None:
            raise MissingRef

        def thunk() -> graphql.GraphQLUnionType:
            types = [exec_thunk(res, non_null=False) for res in results]
            if name is None:
                assert self.union_ref_factory is not None
                computed_name = self.union_ref_factory([t.name for t in types])
            else:
                computed_name = name
            return graphql.GraphQLUnionType(
                computed_name, types, description=description
            )

        return thunk


AwaitableOrNot = Union[Awaitable[T], T]
Subscribe = Callable[..., AwaitableOrNot[AsyncIterable]]

async_iterable_origins = set(map(get_origin, (AsyncIterable[Any], AsyncIterator[Any])))

_fake_type = cast(type, ...)


def graphql_schema(
    *,
    query: Iterable[Callable] = (),
    mutation: Iterable[Callable] = (),
    subscription: Iterable[Union[Subscribe, Tuple[Subscribe, Callable]]] = (),
    types: Iterable[Type] = (),
    aliaser: Aliaser = to_camel_case,
    id_types: Union[Collection[AnyType], IdPredicate] = None,
    error_as_null: bool = True,
    generic_ref_factory: GenericRefFactory = None,
    union_ref_factory: UnionRefFactory = None,
    directives: Optional[Collection[graphql.GraphQLDirective]] = None,
    description: Optional[str] = None,
    extensions: Optional[Dict[str, Any]] = None,
) -> graphql.GraphQLSchema:
    def operation_resolver(operation: Callable, *, skip_first=False) -> Resolver:
        if skip_first:
            wrapper = operation
        else:

            def wrapper(_, *args, **kwargs):
                return operation(*args, **kwargs)

        parameters = resolver_parameters(operation, skip_first=skip_first)
        return Resolver(operation, wrapper, parameters)

    query_fields: List[ObjectField] = []
    mutation_fields: List[ObjectField] = []
    subscription_fields: List[ObjectField] = []
    for operations, fields in [(query, query_fields), (mutation, mutation_fields)]:
        for operation in operations:
            resolver = operation_resolver(operation)
            fields.append(
                ObjectField(
                    operation.__name__,
                    wrap_return_type(resolver.return_type, error_as_null),
                    resolve=resolver_resolve(resolver, aliaser, error_as_null),
                    parameters=field_parameters(resolver),
                    schema=get_schema(operation),
                )
            )
    for operation in subscription:  # type: ignore
        resolve: Callable
        if isinstance(operation, tuple):
            operation, event_handler = operation
            name, schema = event_handler.__name__, get_schema(event_handler)
            try:
                resolver = operation_resolver(event_handler, skip_first=True)
            except MissingFirstParameter:
                raise TypeError(
                    "Subscription resolver must have at least one parameter"
                ) from None
            return_type = resolver.return_type
            subscribe = resolver_resolve(
                operation_resolver(operation),
                aliaser,
                error_as_null,
                serialized=False,
            )
            resolve = resolver_resolve(resolver, aliaser, error_as_null)
        else:
            name, schema = operation.__name__, get_schema(operation)
            resolver = operation_resolver(operation)
            if get_origin(resolver.return_type) not in async_iterable_origins:
                raise TypeError(
                    "Subscriptions must return an AsyncIterable/AsyncIterator"
                )
            return_type = get_args(resolver.return_type)[0]
            subscribe = resolver_resolve(
                resolver, aliaser, error_as_null, serialized=False
            )

            def resolve(_, *args, **kwargs):
                return _

        subscription_fields.append(
            ObjectField(
                name,
                wrap_return_type(return_type, error_as_null),
                parameters=field_parameters(resolver),
                resolve=resolve,
                subscribe=subscribe,
                schema=schema,
            )
        )

    is_id = id_types.__contains__ if isinstance(id_types, Collection) else id_types
    builder = OutputSchemaBuilder(
        aliaser, is_id, error_as_null, generic_ref_factory, union_ref_factory
    )

    def root_type(
        name: str, fields: Collection[ObjectField]
    ) -> Optional[graphql.GraphQLObjectType]:
        if not fields:
            return None
        return exec_thunk(builder.object(type(name, (), {}), fields), non_null=False)

    return graphql.GraphQLSchema(
        root_type("Query", query_fields),
        root_type("Mutation", mutation_fields),
        root_type("Subscription", subscription_fields),
        [exec_thunk(builder.visit(cls), non_null=False) for cls in types],
        directives,
        description,
        extensions,
    )



import re
import sys
from base64 import b64decode, b64encode
from collections import deque
from datetime import date, datetime, time
from decimal import Decimal
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from pathlib import (
    Path,
    PosixPath,
    PurePath,
    PurePosixPath,
    PureWindowsPath,
    WindowsPath,
)
from typing import Deque, List, NewType, Type, TypeVar
from uuid import UUID

from apischema.conversions.converters import deserializer, serializer
from apischema.json_schema.schema import schema


def as_str(cls: Type, format: str = None):
    str_type = schema(format=format)(NewType(cls.__name__.capitalize(), str))
    deserializer(cls, str_type, cls)
    serializer(str, cls, str_type)


T = TypeVar("T")


# =================== bytes =====================

deserializer(b64decode, str, bytes)


@serializer
def to_base64(b: bytes) -> str:
    return b64encode(b).decode()


# ================ collections ==================

deserializer(deque, List[T], Deque[T])
serializer(list, Deque[T], List[T])
if sys.version_info < (3, 7):
    deserializer(deque, List, deque)
    serializer(list, deque, List)


# ================== datetime ===================

if sys.version_info >= (3, 7):  # pragma: no cover
    for cls, format in [(date, "date"), (datetime, "date-time"), (time, "time")]:
        str_type = schema(format=format)(NewType(cls.__name__.capitalize(), str))
        deserializer(cls.fromisoformat, str_type, cls)  # type: ignore
        serializer(cls.isoformat, cls, str_type)  # type: ignore

# ================== decimal ====================

deserializer(Decimal, float, Decimal)
serializer(float, Decimal, float)

# ================= ipaddress ===================

for cls in (IPv4Address, IPv4Interface, IPv4Network):
    as_str(cls, "ipv4")
for cls in (IPv6Address, IPv6Interface, IPv6Network):
    as_str(cls, "ipv6")

# ==================== path =====================

for cls in (PurePath, PurePosixPath, PureWindowsPath, Path, PosixPath, WindowsPath):
    as_str(cls)

# =================== pattern ===================

Pattern = type(re.compile(r""))
deserializer(re.compile, str, Pattern)
serializer(lambda p: p.pattern, Pattern, str)

# ==================== uuid =====================

as_str(UUID, "uuid")



from dataclasses import is_dataclass
from enum import Enum
from typing import Any, Awaitable, Callable, Collection, Dict, Mapping, Optional

import graphql

from apischema.aliases import Aliaser
from apischema.conversions import Conversions
from apischema.conversions.dataclass_model import DataclassModelWrapper
from apischema.conversions.visitor import SerializationVisitor
from apischema.deserialization import deserialize
from apischema.resolvers import Resolver
from apischema.serialization import (
    COLLECTION_TYPE_SET,
    MAPPING_TYPE_SET,
    PRIMITIVE_TYPES_SET,
    serialize,
)
from apischema.types import PRIMITIVE_TYPES
from apischema.validation.errors import ValidationError
from apischema.visitor import Unsupported


def partial_serialize(
    obj: Any,
    *,
    conversions: Conversions = None,
    aliaser: Aliaser = None,
) -> Any:
    assert aliaser is not None
    cls = obj.__class__
    if cls in PRIMITIVE_TYPES_SET:
        return obj
    if cls in COLLECTION_TYPE_SET:
        return [
            partial_serialize(elt, conversions=conversions, aliaser=aliaser)
            for elt in obj
        ]
    if cls in MAPPING_TYPE_SET:
        return serialize(
            obj, conversions=conversions, aliaser=aliaser, exclude_unset=False
        )
    target = None
    if conversions is not None:
        try:
            target = conversions[cls]
        except KeyError:
            pass
    conversion = SerializationVisitor._is_conversion(cls, target)
    if conversion is not None:
        _, (converter, sub_conversions) = conversion
        if isinstance(target, DataclassModelWrapper):
            return obj
        return partial_serialize(
            converter(obj), conversions=sub_conversions, aliaser=aliaser
        )
    if is_dataclass(cls):
        return obj
    if issubclass(cls, Enum):
        return serialize(obj.value, aliaser=aliaser, exclude_unset=False)
    if isinstance(obj, PRIMITIVE_TYPES):
        return obj
    if isinstance(obj, Mapping):
        return serialize(obj, aliaser=aliaser, exclude_unset=False)
    if isinstance(obj, Collection):
        return [partial_serialize(elt, aliaser=aliaser) for elt in obj]
    if issubclass(cls, tuple) and hasattr(cls, "_fields"):
        return obj
    raise Unsupported(cls)


INFO_TYPES = {graphql.GraphQLResolveInfo, Optional[graphql.GraphQLResolveInfo]}


def resolver_resolve(
    resolver: Resolver, aliaser: Aliaser, error_as_null: bool, serialized: bool = True
) -> Callable:
    parameters = resolver.parameters
    types = resolver.types
    info_param = next((p.name for p in parameters if types[p.name] in INFO_TYPES), None)
    arg_types = [(p.name, types[p.name]) for p in parameters if p.name != info_param]
    func = resolver.wrapper
    serialize_result: Callable[[Any], Any]
    if not serialized:

        def serialize_result(result):
            return result

    elif resolver.is_async:

        async def serialize_result(result: Awaitable):
            return partial_serialize(
                await result, conversions=resolver.conversions, aliaser=aliaser
            )

    else:

        def serialize_result(result):
            return partial_serialize(
                result, conversions=resolver.conversions, aliaser=aliaser
            )

    def resolve(self, info, **kwargs):
        errors: Dict[str, ValidationError] = {}
        for arg_name, arg_type in arg_types:
            if arg_name in kwargs:
                try:
                    kwargs[arg_name] = deserialize(arg_type, kwargs[arg_name])
                except ValidationError as err:
                    errors[aliaser(arg_name)] = err
        if errors:
            raise TypeError(serialize(ValidationError(children=errors)))
        if info_param:
            kwargs[info_param] = info
        try:
            return serialize_result(func(self, **kwargs))
        except Exception:
            if error_as_null:
                return None
            else:
                raise

    return resolve



__all__ = ["NotNull", "Skip"]
from typing import Any, Iterator, Sequence, TypeVar, Union

from apischema.types import AnyType
from apischema.utils import UndefinedType
from apischema.visitor import Unsupported, Visitor


class Skipped(Exception):
    pass


class _Skip:
    def __call__(self, *, schema_only: bool):
        return SkipSchema if schema_only else self


Skip = _Skip()
SkipSchema = object()

T = TypeVar("T")
try:
    from typing import Annotated  # type: ignore
except ImportError:
    try:
        from typing_extensions import Annotated  # type: ignore
    except ImportError:
        Annotated = None  # type: ignore

if Annotated is not None:
    NotNull = Union[T, Annotated[None, Skip]]
else:

    class _NotNull:
        def __getitem__(self, item):
            raise TypeError("NotNull requires Annotated (PEP 593)")

    NotNull = _NotNull()  # type: ignore


class SkipVisitor(Visitor[Iterator[AnyType]]):
    def __init__(self, schema_only: bool):
        super().__init__()
        self.schema_only = schema_only

    def annotated(self, cls: AnyType, annotations: Sequence[Any]) -> Iterator[AnyType]:
        for annotation in annotations:
            if annotation is Skip or (self.schema_only and annotation is SkipSchema):
                raise Skipped
        return super().annotated(cls, annotations)

    def union(self, alternatives: Sequence[AnyType]) -> Iterator[AnyType]:
        for alt in alternatives:
            try:
                self.visit(alt)
            except Skipped:
                pass
            except (NotImplementedError, Unsupported):
                yield alt
            else:
                raise NotImplementedError()

    def visit(self, cls: AnyType) -> Iterator[AnyType]:
        if cls is UndefinedType:
            raise Skipped
        return super().visit(cls)


def filter_skipped(
    alternatives: Sequence[AnyType], *, schema_only=False
) -> Iterator[AnyType]:
    return SkipVisitor(schema_only).union(alternatives)



__all__ = ["add_serialized", "serialize", "serialized"]

from dataclasses import dataclass, is_dataclass
from enum import Enum
from inspect import Parameter, iscoroutinefunction
from typing import (
    Any,
    Callable,
    Collection,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
)

from apischema import settings
from apischema.aliases import Aliaser
from apischema.cache import cache
from apischema.conversions.dataclass_model import DataclassModelWrapper, get_model
from apischema.conversions.metadata import get_field_conversions
from apischema.conversions.utils import Conversions
from apischema.conversions.visitor import SerializationVisitor
from apischema.dataclass_utils import dataclass_types_and_fields, get_alias
from apischema.fields import FIELDS_SET_ATTR, fields_set
from apischema.metadata.keys import SKIP_METADATA, check_metadata, is_aggregate_field
from apischema.resolvers import (
    Resolver,
    ResolverDescriptor,
    add_resolver,
    get_resolvers,
    resolver,
)
from apischema.types import COLLECTION_TYPES, MAPPING_TYPES, PRIMITIVE_TYPES
from apischema.utils import Undefined, typed_wraps
from apischema.visitor import Unsupported

PRIMITIVE_TYPES_SET = set(PRIMITIVE_TYPES)
COLLECTION_TYPE_SET = set(COLLECTION_TYPES)
MAPPING_TYPE_SET = set(MAPPING_TYPES)

SerializationMethod = Callable[[Any, Callable], Any]


@dataclass
class SerializationField:
    name: str
    method: SerializationMethod


@cache
def serialization_fields(
    cls: Type, aliaser: Aliaser
) -> Tuple[
    Sequence[Tuple[str, SerializationField]],
    Sequence[SerializationField],
    Sequence[Tuple[str, SerializationMethod]],
]:
    types, fields, _ = dataclass_types_and_fields(cls)  # type: ignore
    normal_fields, aggregate_fields, serialized_fields = [], [], []
    for field in fields:
        if SKIP_METADATA in field.metadata:
            continue
        check_metadata(field)
        field_type = types[field.name]
        method: Callable
        conversions = get_field_conversions(field, field_type)
        if conversions is not None:
            if conversions.serializer is None:

                def method(
                    obj: Any,
                    _serialize: Callable,
                    conversions=conversions.serialization,  # type: ignore
                ) -> Any:
                    return _serialize(obj, conversions=conversions)

            else:
                _, (converter, sub_conversions) = conversions.serialization_conversion(
                    field_type
                )

                def method(
                    obj: Any, _serialize: Callable, conversions=sub_conversions
                ) -> Any:
                    return _serialize(converter(obj), conversions=conversions)

        elif field_type in PRIMITIVE_TYPES_SET:
            method = lambda obj, _: obj  # noqa: E731
        else:
            method = lambda obj, _serialize: _serialize(obj)  # noqa: E731
        field2 = SerializationField(field.name, method)
        if is_aggregate_field(field):
            aggregate_fields.append(field2)
        else:
            normal_fields.append((aliaser(get_alias(field)), field2))
    # TODO handle dataclass model feature
    for name, resolver in get_serialized_resolvers(cls).items():  # noqa: F402

        def method(
            obj: Any,
            _serialize: Callable,
            func=resolver.wrapper,
            conversions=resolver.conversions,
        ) -> Any:
            return _serialize(func(obj), conversions=conversions)

        serialized_fields.append((aliaser(name), method))

    return normal_fields, aggregate_fields, serialized_fields


def serialize(
    obj: Any,
    *,
    conversions: Conversions = None,
    aliaser: Aliaser = None,
    exclude_unset: bool = True,
) -> Any:
    if aliaser is None:
        aliaser = settings.aliaser()
    is_conversion = SerializationVisitor._is_conversion

    def _serialize(
        obj: Any,
        *,
        conversions: Optional[Conversions] = None,
    ) -> Any:
        assert aliaser is not None
        cls = obj.__class__
        if cls in PRIMITIVE_TYPES_SET:
            return obj
        if cls in COLLECTION_TYPE_SET:
            return [_serialize(elt, conversions=conversions) for elt in obj]
        if cls in MAPPING_TYPE_SET:
            return {
                _serialize(key, conversions=conversions): _serialize(
                    value, conversions=conversions
                )
                for key, value in obj.items()
            }
        target = None
        if conversions is not None:
            try:
                target = conversions[cls]
            except KeyError:
                pass
        conversion = is_conversion(cls, target)
        if conversion is not None:
            target, (converter, sub_conversions) = conversion
            if isinstance(target, DataclassModelWrapper):
                cls = get_model(target.cls, target.model)
            else:
                # TODO Maybe add exclude_unset parameter to serializers
                return _serialize(converter(obj), conversions=sub_conversions)
        if is_dataclass(cls):
            fields, aggregate_fields, serialized_fields = serialization_fields(
                cls, aliaser
            )
            if exclude_unset and hasattr(obj, FIELDS_SET_ATTR):
                fields_set_ = fields_set(obj)
                fields = [(a, f) for (a, f) in fields if f.name in fields_set_]
                aggregate_fields = [
                    f for f in aggregate_fields if f.name in fields_set_
                ]
            result = {}
            # properties before normal fields to avoid overloading a field with property
            for field in aggregate_fields:
                attr = getattr(obj, field.name)
                result.update(field.method(attr, _serialize))  # type: ignore
            for alias, field in fields:
                attr = getattr(obj, field.name)
                if attr is not Undefined:
                    result[alias] = field.method(attr, _serialize)  # type: ignore
            for alias, method in serialized_fields:
                res = method(obj, _serialize)
                if res is not Undefined:
                    result[alias] = res
            return result
        if obj is Undefined:
            raise Unsupported(cls)
        if issubclass(cls, Enum):
            return _serialize(obj.value)
        if isinstance(obj, PRIMITIVE_TYPES):
            return obj
        if isinstance(obj, Mapping):
            return {_serialize(key): _serialize(value) for key, value in obj.items()}
        if isinstance(obj, Collection):
            return [_serialize(elt) for elt in obj]
        if issubclass(cls, tuple) and hasattr(cls, "_fields"):
            result = {}
            for field_name in obj._fields:
                attr = getattr(obj, field_name)
                if attr is not Undefined:
                    result[aliaser(field_name)] = attr
            return result
        raise Unsupported(cls)

    return _serialize(obj, conversions=conversions)


def has_parameter_without_default(resolver) -> bool:
    return any(arg.default is Parameter.empty for arg in resolver.parameters)


def can_be_serialized(resolver: Resolver) -> bool:
    return not has_parameter_without_default(resolver) and not resolver.is_async


def check_serialized(cls: Type, name: str):
    resolver = get_resolvers(cls)[name]
    if has_parameter_without_default(resolver):
        raise TypeError(f"{resolver.func} cannot have parameter without default")
    try:
        if resolver.is_async:
            raise TypeError(f"async {resolver.func} cannot be serialized")
    except Exception:  # get_type_hints can fail
        if iscoroutinefunction(resolver.func):
            raise TypeError(f"coroutine {resolver.func} cannot be serialized")


class SerializedDescriptor:
    def __init__(self, resolver_desc: ResolverDescriptor):
        self.resolver_desc = resolver_desc

    def __set_name__(self, owner, name):
        self.resolver_desc.__set_name__(owner, name)
        check_serialized(owner, self.resolver_desc.name or name)


def _serialized(*args, **kwargs):
    result = resolver(*args, **kwargs)
    if isinstance(result, ResolverDescriptor):
        return SerializedDescriptor(result)
    else:
        return lambda method: SerializedDescriptor(result(method))


serialized = typed_wraps(resolver)(_serialized)


def _add_serialized(cls, name=None, **kwargs):
    def decorator(func):
        result = add_resolver(cls, name=name, **kwargs)(func)
        check_serialized(cls, name or func.__name__)
        return result

    return decorator


add_serialized = typed_wraps(add_resolver)(_add_serialized)


def get_serialized_resolvers(cls: Type) -> Mapping[str, Resolver]:
    return {
        name: res for name, res in get_resolvers(cls).items() if can_be_serialized(res)
    }
